"""A copy of Python's rlcompleter module that removes all interaction
with the readline module.  This is a trivial modification of rlcompleter.py
from Python 3.9, which is licensed under the Python Software Foundation
License Version 2.

This plainCompleter module is designed for use in Regina's GUI, which uses
Python subinterpreters.  The readline module does not play well with
subinterpreters (due to the way it interacts with the GIL), and in some
settings this results in a completely reproducible deadlock.  Such settings
include openSUSE Leap 15.2, and macOS 11.0 using the Python3 that is shipped
with Xcode.

This plainCompleter module still offers a Completer class (which is all
that the GUI needs), but it does not offer full readline functionality
at the command line.  Crucially, importing this module will never attempt
to import the readline module as a side-effect.

Python's own rlcompleter documentation continues from here.

The completer completes keywords, built-ins and globals in a selectable
namespace (which defaults to __main__); when completing NAME.NAME..., it
evaluates (!) the expression up to the last dot and completes its attributes.

It's very cool to do "import sys" type "sys.", hit the completion key (twice),
and see the list of names defined by the sys module!

Tip: to use the tab key as the completion key, call

    readline.parse_and_bind("tab: complete")

Notes:

- Exceptions raised by the completer function are *ignored* (and generally cause
  the completion to fail).  This is a feature -- since readline sets the tty
  device in raw (or cbreak) mode, printing a traceback wouldn't work well
  without some complicated hoopla to save, reset and restore the tty state.

- The evaluation of the NAME.NAME... form may cause arbitrary application
  defined code to be executed if an object with a __getattr__ hook is found.
  Since it is the responsibility of the application (or the user) to enable this
  feature, I consider this an acceptable risk.  More complicated expressions
  (e.g. function calls or indexing operations) are *not* evaluated.

- When the original stdin is not a tty device, GNU readline is never
  used, and this module (and the readline module) are silently inactive.

"""

import atexit
import __main__

# The original rlcompleter simply imports the builtins module.
# However, we need to maintain compatibility with python 2 also.
try:
    import builtins
except:
    import __builtin__ as builtins

__all__ = ["Completer"]

class Completer:
    def __init__(self, namespace = None):
        """Create a new completer for the command line.

        Completer([namespace]) -> completer instance.

        If unspecified, the default namespace where completions are performed
        is __main__ (technically, __main__.__dict__). Namespaces should be
        given as dictionaries.

        Completer instances should be used as the completion mechanism of
        readline via the set_completer() call:

        readline.set_completer(Completer(my_namespace).complete)
        """

        if namespace and not isinstance(namespace, dict):
            raise TypeError('namespace must be a dictionary')

        # Don't bind to namespace quite yet, but flag whether the user wants a
        # specific namespace or to use __main__.__dict__. This will allow us
        # to bind to __main__.__dict__ at completion time, not now.
        if namespace is None:
            self.use_main_ns = 1
        else:
            self.use_main_ns = 0
            self.namespace = namespace

    def complete(self, text, state):
        """Return the next possible completion for 'text'.

        This is called successively with state == 0, 1, 2, ... until it
        returns None.  The completion should begin with 'text'.

        """
        if self.use_main_ns:
            self.namespace = __main__.__dict__

        if not text.strip():
            if state == 0:
                if _readline_available:
                    readline.insert_text('\t')
                    readline.redisplay()
                    return ''
                else:
                    return '\t'
            else:
                return None

        if state == 0:
            if "." in text:
                self.matches = self.attr_matches(text)
            else:
                self.matches = self.global_matches(text)
        try:
            return self.matches[state]
        except IndexError:
            return None

    def _callable_postfix(self, val, word):
        if callable(val):
            word = word + "("
        return word

    def global_matches(self, text):
        """Compute matches when text is a simple name.

        Return a list of all keywords, built-in functions and names currently
        defined in self.namespace that match.

        """
        import keyword
        matches = []
        seen = {"__builtins__"}
        n = len(text)
        for word in keyword.kwlist:
            if word[:n] == text:
                seen.add(word)
                if word in {'finally', 'try'}:
                    word = word + ':'
                elif word not in {'False', 'None', 'True',
                                  'break', 'continue', 'pass',
                                  'else'}:
                    word = word + ' '
                matches.append(word)
        for nspace in [self.namespace, builtins.__dict__]:
            for word, val in nspace.items():
                if word[:n] == text and word not in seen:
                    seen.add(word)
                    matches.append(self._callable_postfix(val, word))
        return matches

    def attr_matches(self, text):
        """Compute matches when text contains a dot.

        Assuming the text is of the form NAME.NAME....[NAME], and is
        evaluable in self.namespace, it will be evaluated and its attributes
        (as revealed by dir()) are used as possible completions.  (For class
        instances, class members are also considered.)

        WARNING: this can still invoke arbitrary C code, if an object
        with a __getattr__ hook is evaluated.

        """
        import re
        m = re.match(r"(\w+(\.\w+)*)\.(\w*)", text)
        if not m:
            return []
        expr, attr = m.group(1, 3)
        try:
            thisobject = eval(expr, self.namespace)
        except Exception:
            return []

        # get the content of the object, except __builtins__
        words = set(dir(thisobject))
        words.discard("__builtins__")

        if hasattr(thisobject, '__class__'):
            words.add('__class__')
            words.update(get_class_members(thisobject.__class__))
        matches = []
        n = len(attr)
        if attr == '':
            noprefix = '_'
        elif attr == '_':
            noprefix = '__'
        else:
            noprefix = None
        while True:
            for word in words:
                if (word[:n] == attr and
                    not (noprefix and word[:n+1] == noprefix)):
                    match = "%s.%s" % (expr, word)
                    try:
                        val = getattr(thisobject, word)
                    except Exception:
                        pass  # Include even if attribute not set
                    else:
                        match = self._callable_postfix(val, match)
                    matches.append(match)
            if matches or not noprefix:
                break
            if noprefix == '_':
                noprefix = '__'
            else:
                noprefix = None
        matches.sort()
        return matches

def get_class_members(klass):
    ret = dir(klass)
    if hasattr(klass,'__bases__'):
        for base in klass.__bases__:
            ret = ret + get_class_members(base)
    return ret

# Regina's edits are here.
# Instead of trying to import the readline module, we just ignore it
# and treat this module as a standalone completion tool.
_readline_available = False



################################
#
#  Sample Python Script
#
#  Illustrates progress reporting during long operations.
#
#  See the file "progress.session" for the results of running this script.
#
################################

import threading
import time

# Create an 18-tetrahedron triangulation of a knot complement with real
# boundary faces (not an ideal vertex).  The knot is L106003 from the
# knot/link census.  We used Regina to truncate the ideal vertex, and
# then copied the isomorphism signature so that we can reconstruct the
# triangulation here.
sig = 'sfLfvQvwwMQQQccjghjkmqlonrnrqpqrnsnksaisnrobocksks'
tri = Triangulation3(sig)
print(tri)

# Create a progress tracker to use during the normal surface enumeration.
# This will report the state of progress while the enumeration runs in
# the background.
tracker = ProgressTracker()

# Start the normal surface enumeration in a new thread.
surfaces = None
def run():
    global surfaces, tracker
    surfaces = NormalSurfaces(tri, NormalCoords.Standard, NormalList.Vertex,
        NormalAlg.Default, tracker)

thread = threading.Thread(target = run)
thread.start()

# At this point the enumeration is up and running.
# Output a progress report every quarter-second until it finishes.
while not tracker.isFinished():
    print('Progress:', tracker.percent(), '%')
    time.sleep(0.25)


# The surface enumeration is now complete.
thread.join()
print(surfaces)




# Regina - A Normal Surface Theory Calculator
# Python Module Initialisation
#
# Copyright (c) 2003-2025, Ben Burton
# For further details contact Ben Burton (bab@debian.org).
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# As an exception, when this program is distributed through (i) the
# App Store by Apple Inc.; (ii) the Mac App Store by Apple Inc.; or
# (iii) Google Play by Google Inc., then that store may impose any
# digital rights management, device limits and/or redistribution
# restrictions that are required by its terms of service.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import sys, os
from . import engine
from .engine import *

if 'sage' in sys.modules:
    # Typing "from regina import *" should not override certain
    # names. This always applies to the built-in "open".
    #
    # Also: the Sage preparser adds "Integer" and relies on the name
    # being bound to a Sage integer. If we override it, Sage
    # breaks in multiple ways.
    #
    # We skip names_to_avoid in __all__.
    names_to_avoid = set(['open', 'Integer'])
    __all__ = (
        [ name for name in engine.__dict__.keys()
          if not name in names_to_avoid and not name.startswith('_') ] +
        [ 'reginaSetup' ])

    # All additional sage-related setup should be placed within sageSetup.py.
    from . import sageSetup
    del sageSetup
else:
    # Typing "from regina import *" is not supposed to override the
    # built-in "open".  To achieve this, we skip it in __all__.
    __all__ = (
        [ name for name in engine.__dict__.keys()
          if name != 'open' and not name.startswith('_') ] +
        [ 'reginaSetup' ])

def reginaSetup(quiet = False, readline = True, banner = False,
                snappyPath = True, namespace = None, builtinOpen = True):
    """
    Initialise a Regina python session.

    Arguments:
        quiet      : If true, suppress informative messages that would
                     otherwise be written to standard output.  If an error
                     occurs, details of the error will be written regardless.
        readline   : If true, attempt to enable tab completion.
        snappyPath : Applies to platforms where SnapPy might not be installed
                     on the python path (e.g., macOS users with the SnapPy app
                     bundle).  If true, this setup routine will (i) attempt
                     to locate SnapPy, and (ii) if successful, extend sys.path
                     to include the location of SnapPy's python module and its
                     dependencies.
        banner     : If true, print a welcome banner to standard output.
        namespace  : The global namespace in which the start-up library scripts
                     (if any) will be executed.  This may be None, in which
                     case the caller's global namespace will be used.
        builtinOpen: If true, sets "open" in the given namespace to Python's
                     builtin open() function.  This is used to work around the
                     problem whereby "from regina import *", overrides Python's
                     open() function with Regina's.  You will still be able to
                     access Regina's open() function by calling regina.open().
                     If the namespace argument above is None, then this option
                     has no effect.
    """

    if readline:
        # Enable tab completion through readline, if we can.
        try:
            import rlcompleter, readline
            # readline by default completes an empty string, whereas if 
            # we press tab we want to insert an actual tab character, 
            # so we have our own completion function.

            __internal_python_completer = readline.get_completer()
            def regina_completer(text, state):
              if not text:
                return ('\t', None)[state]
              else:
                return __internal_python_completer(text, state)
            readline.set_completer(regina_completer)

            if 'libedit' in readline.__doc__:
                # Some systems work with libedit, not libreadline, which
                # supports a different set of commands.
                readline.parse_and_bind('bind ^I rl_complete')
            else:
                readline.parse_and_bind('tab: complete')
        except:
            pass

    if snappyPath:
        # For the time being, only find SnapPy on macOS.
        if sys.platform == 'darwin':
            if sys.version_info[:2] == (3, 8):
                # Ask macOS where SnapPy lives.
                import subprocess
                try:
                    app = subprocess.check_output(['mdfind',
                        'kMDItemDisplayName==SnapPy&&kMDItemKind==Application'])
                    if app:
                        app = app.strip().split('\n')[0]
                except:
                    app = None

                if not app:
                    app = '/Applications/SnapPy.app'
                snappyLib = app + '/Contents/Resources/lib/python3.8'
                snappyZip = app + '/Contents/Resources/lib/python38.zip'
                if os.path.exists(snappyLib):
                    sys.path.append(snappyLib)
                    sys.path.append(snappyLib + '/site-packages.zip')
                    sys.path.append(snappyLib + '/lib-dynload')
                if os.path.exists(snappyZip):
                    sys.path.append(snappyZip)

    if banner:
        print(engine.welcome())

    if builtinOpen and namespace:
        namespace['open'] = __builtins__['open']

def __execScript(namespace = None):
    """
    For internal use by regina-python.
    Executes a given python script.

    The filename of the script should be in sys.argv[1], and
    any arguments to the script should be in sys.argv[2:].
    However, ipython sets things up a little differently (it includes two
    additional arguments), and we attempt to compensate for this here.

    SIDE-EFFECT: sys.argv will be truncated to include the script and
    its arguments only (i.e., sys.argv[0] will be removed).

    Arguments:
        namespace : The global namespace in which the script will be executed.
                    This may be None, in which case the caller's global
                    namespace will be used.
    """

    try:
        __IPYTHON__
        # Although python sets sys.argv = [ '-c', script, arg, ... ],
        # ipython sets sys.argv as the full original command line:
        #   [ ipython, '-c', command, script, arg, ... ], or
        #   [ ipython, '-i', '-c', command, script, arg, ... ].
        # Repair things here.
        if len(sys.argv) >= 3 and sys.argv[1] == '-c':
            sys.argv = [ '-c' ] + sys.argv[3:]
        elif len(sys.argv) >= 4 and sys.argv[1:3] == ['-i', '-c']:
            sys.argv = [ '-c' ] + sys.argv[4:]
    except:
        pass

    script = __builtins__['open'](sys.argv[1]).read()
    sys.argv = sys.argv[1:]
    if namespace:
        exec(script, namespace)
    else:
        exec(script)




################################
#
#  Sample Python Script
#
#  Illustrates different queries and actions on a 3-manifold triangulation
#  and its normal surfaces.
#
#  See the file "triangulation.session" for the results of running this
#  script.
#
################################

# Create a new (3,4,7) layered solid torus.  This is a 3-tetrahedron
# triangulation of a solid torus.
t = Example3.lst(3, 4)
print(t)

# Print the full skeleton of the triangulation.
print(t.detail())

# Calculate some algebraic properties of the triangulation.
print(t.homology())
print(t.homologyBdry())

# Test for 0-efficiency, which asks Regina to search for certain types
# of normal surfaces.
print(t.isZeroEfficient())

# Make our own list of vertex normal surfaces in standard coordinates.
surfaces = NormalSurfaces(t, NormalCoords.Standard)

# Print the full list of vertex normal surfaces.
print(surfaces.detail())

# Print the Euler characteristic and orientability of each surface.
for s in surfaces:
    print("Chi =", s.eulerChar(), "; Or =", s.isOrientable())


# List all surfaces with more than one quad in the first tetrahedron.
for s in surfaces:
    if s.quads(0,0) + s.quads(0,1) + s.quads(0,2) > 1:
        print(s)




################################
#
#  Sample Python Script
#
#  Illustrates the traversal and manipulation of an entire packet tree.
#
#  See the file "tree.session" for the results of running this script.
#
################################

# Recreate the original SnapPea census of cusped hyperbolic manifolds
# triangulated by at most 5 tetrahedra.
#
# Since we are building a packet tree, we need to use PacketOfTriangulation3,
# not the plain type Triangulation3 (which is not a packet type).
census = Container()
for i in range(415):
    mfd = SnapPeaCensusManifold(SnapPeaCensusManifold.SEC_5, i)
    census.append(make_packet(mfd.construct(), mfd.name()))

# The triangulations are now all children of the "census" container.
# Remove all triangulations with more than two tetrahedra.
#
# Since we are deleting children, we step through the children manually
# instead of just iterating over children().
tri = census.firstChild()
while tri != None:
    next = tri.nextSibling()
    if tri.size() > 2:
        tri.makeOrphan()
    tri = next

# Print the homology of each remaining triangulation.
# This time we are not adding or removing children, so we can just iterate.
for tri in census.children():
    print(tri.label() + ":", tri.homology())

